import os
import shutil
import argparse
import subprocess
from ..templates import TEMPLATES
from ..command import Command
from ..changelog import Changelog
from ..template import setup_project


def is_git_initialized():
    try:
        subprocess.check_call(['git', 'status'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return True
    except subprocess.CalledProcessError:
        return False


def commit_configuration():
    try:
        subprocess.call(['git', 'add', 'docs/', '.github/', 'setup.cfg'])
        subprocess.check_call(['git', 'commit', '-m', 'Initialize devops configuration files'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        print('Devops configuration files added to git')
    except subprocess.CalledProcessError:
        pass


def path(string):
    return os.path.abspath(os.path.expanduser(string))


def template(string):
    from ..template import TemplateManifest
    from ..shell import format_table
    manifest = TemplateManifest(os.path.expanduser("~/.git-templates"))
    if string not in manifest.templates:
        print("Template '%s' not found...\nPlease choose one of the following, or install a new template:\n\n%s"%(string, format_table(manifest.templates)))
        exit(0)
    return string


class SetupCommand(metaclass=Command):
    def parser():
        parser = argparse.ArgumentParser('git setup')

        parser.add_argument('directory', type=path, default=".")
        parser.add_argument('--template', type=template, help='Template used to initiate project structure')
        parser.add_argument('-v', '--verbose', action='store_true', help='Display information about created files')

        return parser

    def run(*, directory=".", template=None, verbose=False):
        setup_project(template=template, destination=directory)



import argparse
import subprocess
from ..command import Command


class SaveCommand(metaclass=Command):
    def parser():
        parser = argparse.ArgumentParser()
        parser.add_argument("description", nargs="?", help="Commit message")
        return parser

    def run(*, description=None):
        if not description:
            description = input("description: ")
        
        subprocess.call(['git', 'add', '.'])
        subprocess.call(['git', 'commit', '-m', description])
    


import argparse
import subprocess
from ..command import Command
from ..constants import BRANCHES_PREFIXES


class WipCommand(metaclass=Command):
    def parser():
        parser = argparse.ArgumentParser("git wip")
        parser.add_argument("branch_type", type=str, choices=BRANCHES_PREFIXES.keys(), help="Category of the changes done on this branch")
        parser.add_argument("description", nargs='?', type=str, help="Description of the work done on this branch")
        return parser

    def run(*, branch_type, description):
        """
        Create a typed branch to work on a feature/bugfix/...
        """        
        branch_description = description or input('branch description: ')
        branch_name = branch_type + "/" + branch_description.replace(' ', '_').lower()
        
        subprocess.call(['git', 'checkout', 'main'])
        subprocess.call(['git', 'pull', '--rebase', 'origin', 'main'])
        subprocess.call(['git', 'checkout', '-b', branch_name])
        subprocess.call(['git', 'push', '-u', 'origin', branch_name])



import re
import os
from datetime import datetime


def sentence(text):
    return text[0].upper() + text[1:]


class SectionChange:
    def __init__(self, description, indent=1):
        self.description = description
        self.indent = indent


class ReleaseSection:
    def __init__(self, *, title, changes=None):
        self.title = title
        self.changes = changes or []

    def add_change(self, change, indent=1):
        self.changes.append(SectionChange(change, indent=indent))

    def contains_change(self, searched_change):
        for change in self.changes:
            if change.description == searched_change.description:
                return True
        return False

    def difference(self, section):
        if section is None:
            return None
        
        if self.title != section.title:
            raise ValueError("Can't compare different sections...")

        changes = []
        for change in self.changes:
            if not section.contains_change(change):
                changes.append(change)
        
        if len(changes) == 0:
            return None
        
        return ReleaseSection(title=self.title, changes=changes)
            

    def render(self):
        rendered = "**%s**\n\n"%sentence(self.title)
        
        for change in self.changes:
            rendered += "\t"*change.indent + "- %s\n"%sentence(change.description)

        rendered += "\n"

        return rendered


class Release:
    """
    Abstraction used to document release in a changelog.
    """
    def __init__(self, *, version, date, sections=None):
        self.version = version
        self.date = date
        self.sections = sections or []


    def add_section(self, section):
        """
        Add a section at the end of this release.
        """
        self.sections.append(section)


    def get_section(self, title):
        """
        Extract a section from this release given its title.
        """
        for section in self.sections:
            if section.title.lower() == title.lower():
                return section
        
        section = ReleaseSection(title=title)
        self.sections.append(section)
        return section


    def difference(self, release):
        """
        Workout a release that is the difference between this release and another.
        """
        sections = []
        for section in self.sections:
            existing_section = release.get_section(section.title)
            diff = section.difference(existing_section)
            if diff is not None:
                sections.append(diff)    

        return Release(version=self.version, date=self.date, sections=sections)


    def describe(self):
        """
        Standalone description used to document releases on github.
        """
        rendered = "*Release [%s] - %s*\n\n"%(self.version, datetime.now().strftime("%Y-%m-%d"))

        for section in self.sections:
            rendered += section.render()

        return rendered

    def render(self, *, level=1):
        """
        Return a markdown description of this release.
        """
        rendered = "#"*level + " [%s]"%self.version

        if self.date:
            rendered += " - %s"%self.date

        rendered += "\n\n"

        for section in self.sections:
            rendered += section.render()

        return rendered


class ChangelogParser:
    @classmethod
    def __open(cls, filename):
        try:
            return open(filename)
        except FileNotFoundError:
            with open(filename, 'w') as file:
                file.write('# Changelog\n\n## [Unreleased]\n\n**Structure**\n\n- Initialize project structure')
            print('Changelog initialized in:', filename)
            return open(filename)
    
    @classmethod
    def parse(cls, filename, level=1):
        with cls.__open(filename) as file:
            rows = file.read().split('\n')

        title_pattern = re.compile("^" + level*"#" + "\s+(?P<title>.*changelog.*)", re.IGNORECASE)
        release_pattern = re.compile("^" + (level+1)*"#" + "\s+\[(?P<version>.+)\](?: - (?P<date>\d{4}-\d{2}-\d{2}))?")
        section_pattern = re.compile("\*\*(?P<title>.+)\*\*")
        change_pattern = re.compile("^(?P<indentation>\t*)-\s(?P<description>.*)")

        changelog_start = -1

        title = None
        description = None
        releases = []

        current_release = None
        current_section = None

        for i in range(len(rows)):
            changelog_title_match = title_pattern.match(rows[i])
            release_match = release_pattern.match(rows[i])
            section_match = section_pattern.match(rows[i])
            change_match = change_pattern.match(rows[i])
            
            if changelog_title_match:
                title = changelog_title_match.groupdict()['title']
            elif release_match:
                if current_release:
                    if current_section:
                        current_release.add_section(current_section)
                        current_section = None
                    releases.append(current_release)
                    current_release = None
                release = release_match.groupdict()
                current_release = Release(version=release['version'], date=release.get('date'))
            elif section_match and current_release:
                if current_section:
                    current_release.add_section(current_section)
                    current_section = None

                section = section_match.groupdict()
                current_section = ReleaseSection(title=section['title'])
            elif change_match and current_section:
                change = change_match.groupdict()
                current_section.add_change(change['description'], indent=len(change['indentation']))
        
        if current_release:
            if current_section:
                current_release.add_section(current_section)
            releases.append(current_release)

        return Changelog(title=title, description=description, releases=releases)


class Changelog:
    def __init__(self, *, title, description=None, releases=None):
        self.title = title
        self.description = description
        self.releases = releases or []

    def get_unreleased(self):
        for release in self.releases:
            if release.version.lower() == 'unreleased' or not release.date:
                return release
        
        unreleased = Release(version='Unreleased', date=None)
        self.releases.append(unreleased)
        return unreleased

    def get_release(self, version):
        for release in self.releases:
            if release.version.lower() == version:
                return release

        raise KeyError("Version not found in changelog")

    def delete_release(self, version):
        try:
            release = self.get_release(version)
            self.releases.remove(release)
        except KeyError:
            print("[Warning] Can't delete version %s from changelog"%version)

        try:
            os.remove("docs/releases/%s.md"%version)
        except FileNotFoundError:
            print("[Warning] Can't delete 'docs/releases/%s.md': file not found"%version)


    def add_change(self, section, description, changes=None):
        unreleased = self.get_unreleased()
        section = unreleased.get_section(section)
        section.add_change(description, indent=0)

        changes = changes or []
        for change in changes:
            section.add_change(change, indent=1)

    def render(self, *, level=1):
        rendered = "#"*level + " %s"%self.title + "\n\n"

        if self.description:
            rendered += self.description + "\n\n"

        for release in self.releases:
            rendered += release.render(level=level+1)
        
        return rendered

    def save(self, filename):
        with open(filename, 'w') as file:
            file.write(self.render())

    def create_release(self, version, release_file=None):
        release = self.get_unreleased()
        release.version = version
        release.date = datetime.now().strftime('%Y-%m-%d')
        if release_file:
            with open(release_file, 'w') as release_file:
                release_file.write(release.describe())

    @classmethod
    def parse(self, filename):
        return ChangelogParser.parse(filename)



import argparse
import subprocess
from ..menu import Menu
from ..command import Command
from ..git import get_local_branches, get_remote_branches, get_current_branch, delete_branch


def select_branches():
    branches = set(get_local_branches() + get_remote_branches())
    try:
        branches.remove("main")
    except KeyError:
        pass

    select_branches_menu = Menu(choices=list(branches))
    return select_branches_menu.prompt("Select branches to delete")


class CancelCommand(metaclass=Command):
    def parser():
        parser = argparse.ArgumentParser("git cancel")
        parser.add_argument("-i", "--interactive", action='store_true', help="Prompt names of multiple branches to delete")
        return parser

    def run(interactive=False):
        """
        Delete current branch from local and remote
        """
        current_branch = get_current_branch()

        if interactive:
            branches_to_delete = select_branches()
        else:
            if current_branch == 'main':
                raise Exception("Can't delete branch main...")
            input("Branch '%s' will be deleted. Press [enter] to continue..."%current_branch)
            branches_to_delete = [current_branch]
        
        for unwanted_branch in branches_to_delete:
            if unwanted_branch == current_branch:
                # switch to main branch as you can't cancel a checked out branch
                subprocess.call(['git', 'checkout', 'main'])
            
            delete_branch(unwanted_branch)



import re
import uuid
import argparse
import subprocess
from ..command import Command
from ..constants import BRANCHES_PREFIXES, CATEGORIES_ICONS
from ..changelog import Changelog
from ..git import (
    get_current_branch, merge_squash, get_current_branch_commits,
    ensure_git_initialized, ensure_working_tree_clean, ensure_branch_mergeable
)


class DoneCommand(metaclass=Command):
    def parser():
        parser = argparse.ArgumentParser('git done')
        return parser

    def run():
        ensure_git_initialized()
        #ensure_working_tree_clean()
        ensure_branch_mergeable()
        
        current_branch = get_current_branch()
        branch_type, branch_description = current_branch.split('/')

        change_category = BRANCHES_PREFIXES[branch_type]
        category_icon = CATEGORIES_ICONS[change_category]
        branch_description = branch_description.strip().replace('_', ' ')

        # sort commits from the oldest to the most recent
        commits = list(reversed(get_current_branch_commits()))

        # keep a track of original changelog
        original_changelog = Changelog.parse('docs/changelog.md')

        # parse changelog and add branch changes to [unreleased] appropriate section
        changelog = Changelog.parse('docs/changelog.md')
        changelog.add_change(change_category, branch_description, [commit['message'] for commit in commits])
        changelog.save('docs/changelog.md')

        # open changelog for editing
        subprocess.call(['code', 'docs/changelog.md'])
        input("Edit changelog if required. Then press [enter] to continue...")

        # parse changelog to ensure syntax is still valid after edition
        edited_changelog = Changelog.parse('docs/changelog.md')
        edited_changelog.save('docs/changelog.md')

        # workout changes added to changelog for this branch
        original_unreleased = original_changelog.get_unreleased()
        edited_unreleased = edited_changelog.get_unreleased()

        # Commit all changes
        subprocess.call(['git', 'add', '.'])
        subprocess.call(['git', 'commit', '-m', 'update changelog'])

        # workout squash commit message
        squash_description = category_icon + " " + branch_description

        changes_diff = edited_unreleased.difference(original_unreleased)
        if len(changes_diff.sections) > 0:
            changes = ["- %s"%change.description for change in changes_diff.sections[0].changes if change.indent > 0]
        
            # only add changes to description if more than one change was done
            # to keep description more readable
            if len(changes) > 1:
                squash_description +=  "\n" + "\n".join(changes)
        
        # merge-squash all commits from current branch in main branch
        merge_squash(current_branch, message=squash_description)



import os
import re
import json
import shutil
import requests
import argparse
import subprocess
import importlib.util
from getpass import getpass
from tqdm import tqdm
from jinja2 import Environment, FileSystemLoader, select_autoescape
from .shell import check_output, silent_call
from .git import get_remote_origin, count_changes_from_remote


def load_environment(template_path):
    spec = importlib.util.spec_from_file_location("configure", os.path.join(template_path, "configure.py"))
    configure = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(configure)
    return configure.environment


def setup_project(*, template, destination, verbose=False):
    template_root = os.path.abspath(os.path.join(os.path.expanduser("~/.git-templates"), template))
    templates_path = os.path.join(template_root, 'template')

    print(template_root)
    print(templates_path)

    if not os.path.isdir(templates_path):
        raise FileNotFoundError("template/ folder not found in %s"%template_root)

    environment = load_environment(template_path=template_root)

    templating_engine = Environment(
        loader=FileSystemLoader(searchpath=templates_path),
        autoescape=select_autoescape(['py'])
    )

    # associate template file name to dest file name (name is templated)
    files = []

    for obj in os.walk(templates_path):
        source_folder = obj[0]
        
        # create directories that contains files
        for directory in obj[1]:
            source = os.path.join(source_folder, directory)
            dest = templating_engine.from_string(source.replace(templates_path, destination)).render(**environment)
            os.makedirs(dest, exist_ok=True)

        # create files in current directory
        for filename in obj[2]:
            source = os.path.join(source_folder, filename)
            dest = templating_engine.from_string(source.replace(templates_path, destination)).render(**environment)
            template = templating_engine.get_template(source.replace(templates_path, '').strip("/"))

            with open(dest, 'w') as file:
                file.write(template.render(**environment))

            if verbose:
                print(dest, "created")


class TemplateManifest:
    def __init__(self, templates_path):
        self.templates_path = templates_path
        self.path = os.path.join(templates_path, 'templates.json')
        self.__content = None

    def __open(self):
        try:
            return open(self.path)
        except FileNotFoundError:
            os.makedirs(self.templates_path, exist_ok=True)        
            with open(self.path, 'w') as manifest:
                json.dump({"templates": []}, manifest, indent=4)
            return open(self.path)

    def __save(self):
        content = self.content
        with open(self.path, 'w') as manifest:
            json.dump(content, manifest, indent=4)

    @property
    def content(self):
        if not self.__content:
            self.__content = json.load(self.__open())
        return self.__content

    def add_template(self, template:str):
        if template in self.templates:
            raise FileExistsError("A template named %s already exists in %s"%(template, self.templates_path))
        self.templates.append(template)
        self.__save()

    def remove_template(self, template:str):
        if template in self.templates:
            self.templates.remove(template)
            self.__save()

    @property
    def templates(self):
        return self.content["templates"]


REPO_URL_PATTERN = re.compile("(?P<protocol>\w+)\:\/\/(?:(?P<username>.+)\:(?P<token>.+)@)?(?P<domain>[^/]+)/")

class GitCredentials:
    def __init__(self):
        self.path = os.path.expanduser("~/.git-credentials")
        self.__credentials = None


    def __save(self):
        content = ""
        for credential in self.credentials:
            content += f"{credential['protocol']}://{credential['username']}:{credential['token']}@{credential['domain']}\n"
        
        with open(self.path, "w") as credentials_file:
            credentials_file.write(content)


    def has_credentials(self, *, repository=None, domain=None):
        if (repository is None) == (domain is None):
            raise IOError("Either repository or domain must be passed. (not both)")

        if domain is not None:
            for credential in self.credentials:
                if domain in credential["domain"]:
                    return True
            return False

        if repository is not None:
            match = REPO_URL_PATTERN.match(repository)

            if not match:
                raise ValueError("%s does not look like a valid repository URL."%repository)

            info = match.groupdict()
            if info["username"] and info["token"]:
                return True

        return False

    
    def require_credentials(self, *, repository=None, domain=None):
        
        if not self.has_credentials(repository=repository, domain=domain):
            if repository is not None:
                response = requests.get(repository)
                if response.status_code >= 200 and response.status_code < 300:
                    return "This repo does not requires authentication"
                
                match = REPO_URL_PATTERN.match(repository)

                if not match:
                    raise ValueError("%s does not look like a valid repository URL."%repository)

                scope = match.groupdict()
            else:
                scope = {"domain": domain}
            
            self.authenticate(**scope)


    def authenticate(self, *, domain, username=None, token=None, protocol="https"):
        if self.has_credentials(domain=domain):
            print("Already authenticated")
            return

        if not username or not token:
            print("Authentication required for %s"%domain)
        
        self.credentials.append({
            "protocol": protocol,
            "username": username if username else input("username: "),
            "token": token if token else getpass(),
            "domain": domain
        })
        self.__save()


    @property
    def credentials(self):
        if not self.__credentials:
            credential_pattern = re.compile("(?P<protocol>\w+)\:\/\/(?P<username>.+)\:(?P<token>.+)@(?P<domain>.+)")
            try:
                with open(self.path) as credentials_file:
                    content = credentials_file.read()
                self.__credentials = [m.groupdict() for m in credential_pattern.finditer(content)]
            except FileNotFoundError:
                self.__credentials = {}
        return self.__credentials


def log_update(template, status):
    if status == "updated":
        status_message = "\033[33m[updated]\033[00m  "
    elif status == "unchanged":
        status_message = "\033[32m[unchanged]\033[00m"
    elif status == "failed":
        status_message = "\033[31m[failed]\033[00m   "
    else:
        raise ValueError("Unknown update status received: %s"%status)
    print("%s  %s"%(status_message.ljust(10), template))


class Template:
    def __init__(self):
        self.templates_directory = os.path.expanduser("~/.git-templates")
        self.manifest = TemplateManifest(self.templates_directory)

    def search(self, *, query=None, abspath=False):
        return [
            os.path.join(self.templates_directory, template) if abspath else template
            for template in self.manifest.templates
            if not query or query in template
        ]
    
    def install(self, *, template, origin, user=None, token=None, branch='main'):
        self.manifest.add_template(template)
        template_path = os.path.join(self.templates_directory, template)
        os.makedirs(template_path)
        os.chdir(template_path)

        # retrieve project in .git-templates/ folder
        subprocess.call(['git', 'init'])
        subprocess.call(['git', 'remote', 'add', 'origin', origin])
        subprocess.call(['git', 'checkout', '-b', branch])
        subprocess.call(['git', 'pull', 'origin', branch])
        subprocess.call(['git', 'branch', '--set-upstream-to=origin/%s'%branch, branch])
            

    def update(self, template, verbose=False):        
        template_path = os.path.join(self.templates_directory, template)
        update_target = template

        try:
            os.chdir(template_path)
            if verbose: # workout additionals infos in verbose mode
                update_target += "\t(%s)"%get_remote_origin()
        
            silent_call("git", "fetch")
            if count_changes_from_remote():
                subprocess.check_call(['git', 'pull'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                log_update(update_target, status="updated")
            else:
                log_update(update_target, status="unchanged")
        except Exception:
            log_update(update_target, status="failed")

    def remove(self, template):
        try:
            template_path = os.path.join(self.templates_directory, template)
            self.manifest.remove_template(template)
            shutil.rmtree(template_path)
        except FileNotFoundError as e:
            print("[Warning] %s not found!"%template_path)
        print('template removed:', template)



import os


KEY_SPACE = "KEY_SPACE"
KEY_UP = "KEY_UP"
KEY_DOWN = "KEY_DOWN"
KEY_ENTER = "KEY_ENTER"


def _getkeycode():
    import sys, tty, termios
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(sys.stdin.fileno())
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

    return ord(ch)

def _getkey():
    code = _getkeycode()
    if code == 13:
        return KEY_ENTER
    elif code == 32:
        return KEY_SPACE
    elif code == 27:
        print(_getkeycode())
        if True:
            arrow_code = _getkeycode()
            if arrow_code == 65:
                return KEY_UP
            elif arrow_code == 66:
                return KEY_DOWN
    return None


class Menu:
    def __init__(self, *, choices):
        if not choices or len(choices) == 0:
            raise ValueError("You must provide a non empty list of choices (strings) to Menu")

        self.choices = choices
        self.length = len(choices)
        self.selected = dict((i, False) for i in range(self.length))
        self.cursor = 0

    def clear(self):
        os.system('clear')

    def draw(self, message, multiple=True):
        self.clear()
        print(message)
        for i in range(len(self.choices)):
            
            if multiple:
                cursored_pattern = "[%s]" if self.cursor == i else " %s "
                selected_symbol = "*" if self.selected[i] else " "
                prefix = cursored_pattern % selected_symbol
            else:
                prefix = ">" if self.cursor == i else " "
            
            print(prefix, self.choices[i])

    def prompt(self, message, *, multiple=True):
        char = None
        while char != KEY_ENTER:
            self.draw(message, multiple=multiple)
            char = _getkey()

            if char == KEY_UP:
                self.cursor = (self.cursor - 1) % self.length
            elif char == KEY_DOWN:
                self.cursor = (self.cursor + 1) % self.length
            elif char == KEY_SPACE and multiple:
                self.selected[self.cursor] = not self.selected[self.cursor]
        
        if multiple:
            return [self.choices[i] for i in range(self.length) if self.selected[i]]
        else:
            return self.choices[self.cursor]


if __name__ == '__main__':
    menu_delete_branch = Menu(choices=["feature/useless_feature", "doc/explain_nothing", "improve/nothing_but_stupidity"])
    selection = menu_delete_branch.prompt("Select branch(es) to delete", multiple=True)
    print(selection)



import argparse
from tqdm import tqdm

from ..exceptions import ValidationError
from ..command import Command
from ..template import Template, GitCredentials
from ..shell import display_table


class TemplateCommand(metaclass=Command):
    def parser():
        parser = argparse.ArgumentParser('git template')
        subparsers = parser.add_subparsers(dest="command")

        install_parser = subparsers.add_parser("install", help="install a template from a remote source")
        install_parser.add_argument("alias", help="the name you want to use locally to reference this template")
        install_parser.add_argument("origin", help="url of the repository where target template is hosted")
        install_parser.add_argument("--branch", default="main", help="branch of the repo containing the template (default: main)")
        install_parser.add_argument("--user", help="username used to autenticate if required")
        install_parser.add_argument("--token", help="pass a token or password to authenticate if required")

        update_parser = subparsers.add_parser("update", help="update given templates (default to all).")
        update_parser.add_argument("templates", nargs="*", help="if you pass a list of templates, only these templates will be updated...")
        update_parser.add_argument("-v", "--verbose", action="store_true", help="(verbose) display additional informations...")

        update_parser = subparsers.add_parser("list", help="list installed templates.")
        update_parser.add_argument("search", nargs="?", help="python-style regex that can be used to filter output")

        remove_parser = subparsers.add_parser("rm", help="remove installed templates.")
        remove_parser.add_argument("templates", nargs="+", help="alias(es) of the template(s) to remove")

        return parser

    def validate(args):
        if not args.command:
            raise ValidationError('You must specify a command')

    def run(*, command, alias=None, search=None, branch=None, user=None, token=None, origin=None, templates=None, verbose=False):
        git_template = Template()

        if command == "install":
            # ensure that user can read this repository (=> public repo or proper authentication)
            GitCredentials().require_credentials(repository=origin)
            git_template.install(template=alias, origin=origin, branch=branch)
        elif command == "update":
            templates = templates if templates else git_template.search()
            for template in templates:
                git_template.update(template, verbose=verbose)
        elif command == "list":
            templates = git_template.search(query=search)
            display_table(templates)
        elif command == "rm":
            for template in templates:
                git_template.remove(template)



import re
import argparse
import subprocess
from ..command import Command
from ..changelog import Changelog
from ..constants import CATEGORIES_ICONS
from ..exceptions import ValidationError
from ..git import get_current_branch, ensure_main_branch


# custom type used to parse semver
def version(text):
    try:
        version_pattern = re.compile('^\d+\.\d+\.\d+$')
        if version_pattern.match(text):
            return text
    except Exception:
        pass
    raise ValueError("Not a valid version: %s"%text)


def increment_version(version=None, patch=False, minor=False, major=False):
    incremented_part = 'patch'*patch + 'minor'*minor + 'major'*major

    # an incremented part is always required, even if a version is specified
    if not incremented_part:
        incremented_part = 'patch'
    
    bump_command = ['bump2version']
    if version:
        bump_command.extend(['--new-version', version])
    bump_command.append(incremented_part)

    try:
        subprocess.check_call(bump_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except subprocess.CalledProcessError as e:
        print('Cant increment version, working tree has uncommited changes.')
        exit(1)

    # parse new version from setup.cfg
    with open('setup.cfg') as setup_file:
        rows = setup_file.read().split('\n')
    
    for row in rows:
        if row.startswith('current_version'):
            return row.split('=')[1].strip()
    
    # fail if no version is found
    raise Exception("Version not found in setup.cfg")


def commit_version_files(version):
    subprocess.call(['git', 'add', '.'])
    subprocess.call(['git', 'commit', '-m', '%s Release version %s'%(CATEGORIES_ICONS['Release'], version)])
    subprocess.call(['git', 'tag', version, '-m' 'Release version %s'%version])
    subprocess.call(['git', 'push', '-u', 'origin', 'main', '--follow-tags'])


class ReleaseCommand(metaclass=Command):
    def parser():
        parser = argparse.ArgumentParser()

        group = parser.add_mutually_exclusive_group()
        group.add_argument('--version', type=version, help='Publish a release with given VERSION')
        group.add_argument('--patch', action='store_true', help='Publish a release of type "patch" (auto-increment current version)')
        group.add_argument('--minor', action='store_true', help='Publish a release of type "minor" (auto-increment current version)')
        group.add_argument('--major', action='store_true', help='Publish a release of type "major" (auto-increment current version)')
        
        return parser


    def validate(args):
        if not args.patch and not args.minor and not args.major and not args.version:
            raise ValidationError("At least one argument must be supplied in VERSION, --patch, --minor, --major.")


    def run(version=None, patch=False, minor=False, major=False):
        ensure_main_branch()
        new_version = increment_version(version=version, patch=patch, minor=minor, major=major)
        
        changelog = Changelog.parse('docs/changelog.md')
        release = changelog.get_unreleased()
        changelog.create_release(new_version, release_file="docs/releases/%s.md"%new_version)
        changelog.save('docs/changelog.md')

        commit_version_files(new_version)



from .exceptions import ValidationError


class Command(type):
    def parser(cls):
        """
        Return:
            A parser used to check command format, and return parsed arguments.
        """
        raise NotImplementedError("You must implement a parser for this command")

    def validate(cls, args):
        """
        Implement extra validation for parsed arguments if required...
        """

    def run(cls, **args):
        raise NotImplementedError("You must implement 'run' method for this command")
    
    def __call__(cls):
        parser = cls.parser()
        args = parser.parse_args()

        try:
            cls.validate(args)
        except ValidationError as e:
            parser.print_help()
            print('[Error]', str(e))
            exit(0)
        
        cls.run(**vars(args))



import re
import argparse
import subprocess
from ..git import get_releases
from ..menu import Menu
from ..command import Command
from ..changelog import Changelog
from ..exceptions import ValidationError


def delete_release(version):
    subprocess.call(['git', 'tag', '-d', version])
    subprocess.call(['git', 'push', '--delete', 'origin', version])

    changelog = Changelog.parse('docs/changelog.md')
    changelog.delete_release(version)
    changelog.save('docs/changelog.md')


class UnreleaseCommand(metaclass=Command):
    def parser():
        parser = argparse.ArgumentParser('git unrelease')
        parser.add_argument('version', nargs='?', help='coma separated list of versions to delete')
        parser.add_argument('-i', '--interactive', action='store_true', help="Interactively prompt list of releases to delete.")

        return parser

    def validate(args):
        if (args.interactive == False) == (args.version is None):
            raise ValidationError('Either version, or interactive mode is required. Not both...')

    def run(interactive=False, version=None):        
        if interactive:
            select_releases_menu = Menu(choices=get_releases())
            releases = select_releases_menu.prompt('Select releases to delete...')
            for release in releases:
                delete_release(release)
        else:
            delete_release(version)




BRANCHES_PREFIXES = {
    'bugfix':           'Bugfix',
    'fix':              'Bugfix', 
    'documentation':    'Documentation', 
    'doc':              'Documentation',
    'feature':          'Feature', 
    'structure':        'Structure',
    'devops':           'CI/CD',
    'ci':               'CI/CD',
    'cd':               'CI/CD',
    'enhancement':      'Enhancement',
    'improve':          'Enhancement',
    'improvement':      'Enhancement'
}

CATEGORIES_ICONS = {
    'Bugfix':           ':fire:',
    'Documentation':    ':books:',
    'CI/CD':            ':rocket:',
    'Enhancement':      ':sparkles:',
    'Feature':          ':heavy_plus_sign:',
    'Structure':        ':hammer:',
    'Release':          ':package:'
}



import setuptools


def load_long_description():
    with open('README.md') as readme:
        return readme.read()


setuptools.setup(
    name='git-improved',
    description='Add commands to simplify release and publish operation from Git CLI.',
    version='0.1.1',
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            'git-setup=git_improved.commands.setup:SetupCommand',
            'git-template=git_improved.commands.template:TemplateCommand',

            'git-wip=git_improved.commands.wip:WipCommand',
            'git-save=git_improved.commands.save:SaveCommand',

            'git-cancel=git_improved.commands.cancel:CancelCommand',
            'git-done=git_improved.commands.done:DoneCommand',

            'git-release=git_improved.commands.release:ReleaseCommand',
            'git-unrelease=git_improved.commands.unrelease:UnreleaseCommand'
        ]
    },
    install_requires=[
        'bump2version',
        'jinja2',
        'requests'
    ],
    long_description=load_long_description(),
    long_description_content_type="text/markdown"
)



ACTION_PUBLISH_RELEASE_ON_TAG = """
name: Publish release

on:
  push:
    tags:
      - "*"

jobs:
  create-release:
    name: Create Release
    runs-on: ubuntu-latest
    steps:
      - name: Checkout code
        uses: actions/checkout@v2
    
      - name: Get the version
        id: get_version
        run: echo ::set-output name=VERSION::${GITHUB_REF#refs/tags/}
    
      - name: Create Release
        id: create_release
        uses: actions/create-release@v1
        env:
          GITHUB_TOKEN: ${{ secrets.GITHUB_TOKEN }} # This token is provided by Actions, you do not need to create your own token
        with:
          tag_name: ${{ github.ref }}
          release_name: Release ${{ github.ref }}
          body_path: docs/releases/${{ steps.get_version.outputs.VERSION }}.md
          draft: false
          prerelease: false
"""

BUMPVERSION_CONFIG = """
[bumpversion]
current_version = 0.0.1

# [bumpversion:file:setup.py]
"""


import os
import subprocess


def format_table(items):
    terminal_height, terminal_width = [int(x) for x in os.popen('stty size', 'r').read().split()]

    items = sorted(items)
    max_width = max(len(item) for item in items) + 4

    table_width = terminal_width // max_width
    table_height = len(items) // table_width + 1    
    
    string = ""
    for i in range(table_height):
        for j in range(table_width):
            index = j*table_height+i
            if index < len(items):
                item = items[index]
                string += item.ljust(max_width)
        string += "\n"
    return string


def display_table(items):
    print(format_table(items))


def silent_call(*args):
    subprocess.call(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)


def check_output(*args, rows=False):
    result = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    output = result.stdout.read().decode('utf-8')
    if rows:
        return [row.strip() for row in output.strip().split('\n')]
    else:
        return output.strip()



import re
import subprocess
from .shell import check_output, silent_call
from .constants import BRANCHES_PREFIXES

LOG_COMMIT_PATTERN = re.compile("(?:^|\n)commit\s+(?P<commit_sha>\w+).+\nAuthor:\s+(?P<author>.+)\nDate:\s+(?P<date>.+)\n\n\s+(?P<message>.+)")


def ensure_git_initialized():
    try:
        subprocess.check_call(['git', 'status'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except:
        raise Exception("Not in a git repository.")


def ensure_working_tree_clean():
    result = subprocess.Popen(['git', 'status', '--porcelain'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if result.stdout.read() != b"":
        raise Exception("Working directory is not clean, please commit or stash changes and retry.")


def ensure_main_branch():
    current_branch = get_current_branch()
    if current_branch != 'main':
        raise Exception('Releases can only be deployed from branch "main". Current branch: %s' % current_branch)


def ensure_branch_mergeable():
    try:
        current_branch = get_current_branch()
        branch_prefix, branch_description = current_branch.split('/')
    except ValueError:
        raise Exception("Can't perform magic merge on branch %s"%current_branch)
    
    if branch_prefix not in BRANCHES_PREFIXES:
        raise Exception("Can't perform magic merge on branch of type %s"%branch_prefix)


def count_changes_from_remote():
    current_branch = get_current_branch()
    changes_count = int(check_output("git", "rev-list", "{0}..origin/{0}".format(current_branch), "--count"))
    return changes_count


def get_current_branch():
    result = subprocess.Popen(['git', 'branch', '--show-current'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return result.stdout.read().decode('utf-8').strip()


def get_current_branch_commits():
    result = subprocess.Popen(['git', 'log', 'main..HEAD'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return [m.groupdict() for m in LOG_COMMIT_PATTERN.finditer(result.stdout.read().decode('utf-8'))]


def get_remote_origin():
    return check_output('git', 'remote', 'get-url', 'origin')

def get_local_branches():
    return [branch.split()[-1] for branch in check_output("git", "branch", rows=True)]


def get_remote_branches():
    silent_call("git", "fetch")
    return [branch[7:] for branch in check_output("git", "branch", "--remote", rows=True)]


def get_releases():
    silent_call("git", "fetch")
    return check_output("git", "tag", rows=True)


def delete_branch(branch_name, local=True, remote=True):
    if local:
        subprocess.call(['git', 'branch', '-D', branch_name])
    if remote:
        subprocess.call(['git', 'push', '--delete', 'origin', branch_name])


def merge_squash(merged_branch, message=None):
    if message is None:
        message = "Merge branch %s"%merged_branch
    
    subprocess.call(['git', 'checkout', 'main'])
    subprocess.call(['git', 'pull', '--rebase', 'origin', 'main'])
    subprocess.call(['git', 'merge', '--squash', merged_branch])
    input('Check that everything was ok, then press [enter] to commit and delete WIP branch.')

    subprocess.call(['git', 'commit', '-m', message])
    subprocess.call(['git', 'push', '-u', 'origin', 'main'])
    subprocess.call(['git', 'branch', '-D', merged_branch])
    subprocess.call(['git', 'push', '--delete', 'origin', merged_branch])
